/*
******************************************************************************
** PEXlib Texture Mapping Test Code 
**
** $Source: /src/rcs/Cats/XSystem/pex/txm_utility_code/parser.c,v $
** $Revision: 510.1.200.2 $
** $Date: 94/03/08 17:11:23 $
** $Author: hmgr $
******************************************************************************
*/

/*
 *  THIS IS A PRELIMINARY VERSION!!
 */

/*
 * PARSER.C -- This module parses a parameterfile as input.  All of the 
 *             information contained in the file is loaded in to a single
 *             structure of type Input_Parameters defined in param_struct.h.
 *             The structure is then passed to a texture mapping test. 
 *
 *             Designed by Bradley L. Saunders     Oct. 18, 1993
 *             Modified by Bradley L. Saunders     Dec. 20, 1993
 */

/******************************************************************************/
/*                                                                            */
/*  (c) Copyright Hewlett-Packard Company, 1993, Fort Collins, Colorado       */
/*                                                                            */
/*                            All Rights Reserved                             */
/*                                                                            */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for any purpose and without fee is hereby granted,          */
/*  provided that the above copyright notices appear in all copies and that   */
/*  both the copyright notices and this permission notice appear in           */
/*  supporting documentation, and that the name of Hewlett-Packard not be     */
/*  used in advertising or publicity pertaining to distribution of the        */
/*  software without specific, written prior permission.                      */
/*                                                                            */
/*  HEWLETT-PACKARD MAKES NO WARRANTY OF ANY KIND WITH REGARD TO THIS         */
/*  SOFTWARE, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/*  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  Hewlett-Packard    */
/*  shall not be liable for errors contained herein or direct, indirect,      */
/*  special, incidental or consequential damages in connection with the       */
/*  furnishing, performance or use of this software.                          */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <X11/PEX5/PEXlib.h>
#include <X11/PEX5/PEXHPlib.h>
#include <X11/Xlib.h>
#include "param_struct.h"
#include "prototypes.h"

#define COMMENT_DELIMITER       '#'
#define BYTESPERPIXEL 3
#define INPUTSIZE 120

/* Structures for loading in object information */
struct Point
{
  float x;
  float y;
  float z;
  struct Point *next;
};

struct List
{
  int index;
  struct List *next;
};

struct Facet
{
  struct List *vert;
  struct List *norm;
  struct List *color;
  struct Facet *next;
  int vertex_count;
};

void Get_Keyword_Value(char **file, char *keyword, char *value)
/*
 ******************************************************************************
 **    Description:
 **        This function gets the next keyword and its value from the
 **        parameter file.  If the keyword matches the on sought it
 **        passes back a value otherwise it calls Error().
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char **file - points to the parameter file stream.
 **        char *keyword - stores the keyword.
 **        char *value - a string containing the value associated with
 **                      the keyword.
 **
 **    Output parameters:
 **        A char value is passed back in value.
 **
 **    Entry Conditions:
 **        None.
 **
 **    Exit Conditions:
 **        Calls Error() if the keyword is not found.
 **
 **    Error Conditions:
 **       See Exit Conditions.
 **
 **    Referenced by:
 **       Nearly all functions.
 **
 **    Procedure/Function dependencies:
 **        None.
 **
 ******************************************************************************
 */
{
  char keyin[80];
  char errstr[ERRORSTRINGSIZE];

  sscanf(*file,"%s",keyin);
  if(strcmp(keyword, keyin) != 0)
  {
    sprintf(errstr,"Expecting '%s' : Found '%s'",keyword, keyin);
    Error(errstr);
  }
  *file+=strlen(keyin)+1;
  sscanf(*file,"%s",value);
  *file+=strlen(value)+1;
}

void Get_Value(char **file, char *value)
/*
 ******************************************************************************
 **    Description:
 **        This function gets the next value from the
 **        parameter file. 
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char **file - points to the parameter file stream.
 **        char *value - a string containing the next value.
 **
 **    Output parameters:
 **        A char value is passed back in value.
 **
 **    Entry Conditions:
 **        None.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **       Nearly all functions.
 **
 **    Procedure/Function dependencies:
 **        None.
 **
 ******************************************************************************
 */
{
  sscanf(*file,"%s",value);
  *file+=strlen(value)+1;
}

int Convert_Flags(char *value, char *keyword)
/*
 ******************************************************************************
 **    Description:
 **        This function loads initial information
 **        from the parameter file.
 **
 **    Author:
 **       Brad Saunders                Dec. 20, 1993
 **
 **    Input parameters:
 **        char *value - a pointer to the flag string.
 **        char *keyword - A pointer to the keyword calling this function
 **                        for error reporting.
 **
 **    Output parameters:
 **        Returns a True False Flag.
 **
 **    Entry Conditions:
 **        None.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Load_Do_Flags().
 **        Load_Initial_Information().
 **        Load_Texture_Information().
 **
 **    Procedure/Function dependencies:
 **        None.
 **
 ******************************************************************************
 */
{
  char errstr[ERRORSTRINGSIZE];

  if(strcasecmp(value, "True") == 0)
    return(True);
  else
    if(strcasecmp(value, "False") == 0)
      return(False);
    else
    {
      sprintf(errstr, "Improper flag for '%s'", keyword);
      Error(errstr);
    }
}

void Load_Do_Flags(char **file, Input_Parameters *in)
/*
 ******************************************************************************
 **    Description:
 **        This function loads initial information
 **        from the parameter file.
 **
 **    Author:
 **       Brad Saunders                Dec. 20, 1993
 **
 **    Input parameters:
 **        char *file - a pointer to the input file.
 **        Input_Parameters in - stores the information being read.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  char value[INPUTSIZE];

  Get_Keyword_Value(file, "Do_Custom_Background_Color", value);
  in->Do_Custom_Background_Color = 
    Convert_Flags(value, "Do_Custom_Background_Color");

  Get_Keyword_Value(file, "Do_CreateTM", value);
  in->Do_CreateTM = Convert_Flags(value, "Do_CreateTM");

  Get_Keyword_Value(file, "Do_Call_Coord_Utility", value);
  in->Do_Call_Coord_Utility = Convert_Flags(value, "Do_Call_Coord_Utility");

  Get_Keyword_Value(file, "Do_Use_Custom_View_Point", value);
  in->Do_Use_Custom_View_Point = 
    Convert_Flags(value, "Do_Use_Custom_View_Point");

  Get_Keyword_Value(file, "Do_Use_Custom_Lights", value);
  in->Do_Use_Custom_Lights = Convert_Flags(value, "Do_Use_Custom_Lights");

  Get_Keyword_Value(file, "Do_Use_Depth_Cue", value);
  in->Do_Use_Depth_Cue = Convert_Flags(value, "Do_Use_Depth_Cue");

  Get_Keyword_Value(file, "Do_Use_Coord_Source", value);
  in->Do_Use_Coord_Source = Convert_Flags(value, "Do_Use_Coord_Source");

  Get_Keyword_Value(file, "Do_Use_Composition", value);
  in->Do_Use_Composition = Convert_Flags(value, "Do_Use_Composition");

  Get_Keyword_Value(file, "Do_Use_Sampling_LUT", value);
  in->Do_Use_Sampling_LUT = Convert_Flags(value, "Do_Use_Sampling_LUT");

  Get_Keyword_Value(file, "Do_Use_Binding_LUT", value);
  in->Do_Use_Binding_LUT = Convert_Flags(value, "Do_Use_Binding_LUT");

  Get_Keyword_Value(file, "Do_SetTMPerspectiveCorrection", value);
  in->Do_SetTMPerspectiveCorrection = 
    Convert_Flags(value, "Do_SetTMPerspectiveCorrection");

  Get_Keyword_Value(file, "Do_SetTMSampleFrequency", value);
  in->Do_SetTMSampleFrequency = Convert_Flags(value, "Do_SetTMSampleFrequency");

  Get_Keyword_Value(file, "Do_SetTMResourceHints", value);
  in->Do_SetTMResourceHints = Convert_Flags(value, "Do_SetTMResourceHints");
}

void Load_Initial_Information(char **file, Input_Parameters *in)
/*
 ******************************************************************************
 **    Description:
 **        This function loads initial information
 **        from the parameter file.
 **
 **    Author:
 **       Brad Saunders                Dec. 20, 1993
 **
 **    Input parameters:
 **        char *file - a pointer to the input file.
 **        Input_Parameters in - stores the information being read.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  char value[INPUTSIZE];
  char *end_comments;
  
  /* Get Test Description */
  Get_Value(file, value);
  if(strcmp(value, "Begin_comments") != 0)
    Error("Begin_comments not found");
  end_comments = strstr(*file, "End_comments");
  if(end_comments == NULL)
    Error("End_comments not found"); 
  *end_comments = '\0';
  fprintf(stdout,"%s",*file);
  *file = end_comments + 1;
  Get_Value(file, value);

  /* Get Request Type */
  Get_Keyword_Value(file, "PEXRequestType", value);
  in->req_type = Get_Request_Type(value);

  /* Get Color Specifier forBackground Color */
  Get_Color_Specifier(file, &in->bkgd_color, "Background Color");

  /* Get Interior Style */
  Get_Keyword_Value(file, "Interior_style", value);
  in->style = Convert_Constant(value);

  /* Get BF Interior Style */
  Get_Keyword_Value(file, "BFInterior_style", value);
  in->BFstyle = Convert_Constant(value);

  /* Get Reflection Model */
  Get_Keyword_Value(file, "Reflection", value);
  in->reflec_model = Convert_Constant(value);

  /* Get BF Reflection Model */
  Get_Keyword_Value(file, "BFReflection", value);
  in->BFreflec_model = Convert_Constant(value);

  /* Get Culling Mode */
  Get_Keyword_Value(file, "Culling_mode", value);
  in->culling_mode = Convert_Constant(value);

  /* Parse out values from input file */
  Get_Keyword_Value(file, "Number_of_textures", value);
  in->TEXTURES = atoi(value);
}

void Load_Map(PEXExtTexelArray *texture,
              PEXEnumTypeIndex texel_type,
              char *filename)
/*
 ******************************************************************************
 **    Description:
 **        This functions reads a file containing texture data into 
 **        a data structure that is used by the Create and Filter 
 **        texture functions.
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **
 **    Output parameters:
 **        None. 
 **
 **    Entry Conditions:
 **        1) Assume - that a texture filename exists. 
 **
 **    Exit Conditions:
 **        1) If texture file does not exist then Error() is called.
 **
 **    Error Conditions:
 **        See Exit Conditions.
 **
 **    Referenced by:
 **        Load_Texture_Information().
 **
 **    Procedure/Function dependencies:
 **        Load_Texture_Information().
 **
 ******************************************************************************
 */
{
  char errstr[ERRORSTRINGSIZE];
  unsigned char *buf;
  long texture_size;
  FILE *fin;
  int i;
  int status;
  int filesize;

  fin = fopen(filename,"r");
  if(fin == NULL)
  {
    sprintf(errstr,"Map file - '%s' not found",filename);
    Error(errstr);
  }

  status = fread(&texture->dimension.t0, sizeof(short), 1, fin);
  if(status != 1)
  {
    fclose(fin);
    sprintf(errstr, "Unable to read width dimension in file '%s'", filename);
    Error(errstr);
  }
  status = fread(&texture->dimension.t1, sizeof(short), 1, fin);
  if(status != 1)
  {
    fclose(fin);
    sprintf(errstr, "Unable to read width dimension in file '%s'", filename);
    Error(errstr);
  }
  texture->dimension.t2 = 0;

  fseek(fin, 0L, SEEK_END);     /* crude method for getting size in bytes */
  filesize = ftell(fin);
  fseek(fin, 4L, SEEK_SET);

  if(filesize != ((texture->dimension.t0 * 
                   texture->dimension.t1 *
                   3) +
                  (2 * sizeof(unsigned short))))
  {
    fclose(fin);
    sprintf(errstr, "File '%s' has incorrect length", filename);
    Error(errstr);
  }
         
  texture_size = texture->dimension.t0 * texture->dimension.t1;
  buf = (unsigned char *)malloc(texture_size * BYTESPERPIXEL);

 /* 
  * I read in the map a flip it over to be compatable with PEX's
  * origin 0,0 in the lower left 
  */
  for(i=texture->dimension.t1-1; i>-1; i--)
  {
    fread(&buf[i*texture->dimension.t0*BYTESPERPIXEL], 
          BYTESPERPIXEL, 
          texture->dimension.t0, 
          fin);
      
  }
  fclose(fin);

 /*
  * Allocate space to hold the texture and fill in the structure 
  */
  switch(texel_type)
  {
  case (PEXEnumTypeIndex)PEXExtTexelRGBFloat:
    texture->array.rgb =
      (PEXExtTexelRGB *)malloc(texture_size * sizeof(PEXExtTexelRGB));
    for(i=0; i<texture_size; i++)
    {
      texture->array.rgb[i].red   = 
        (float)buf[i * BYTESPERPIXEL]     / 255.0;
      texture->array.rgb[i].green = 
        (float)buf[i * BYTESPERPIXEL + 1] / 255.0;
      texture->array.rgb[i].blue  = 
        (float)buf[i * BYTESPERPIXEL + 2] / 255.0;
    }
  break;

  case (PEXEnumTypeIndex)PEXExtTexelRGBAlphaFloat:
    texture->array.rgb_alpha =
      (PEXExtTexelRGBAlpha *)malloc(texture_size * sizeof(PEXExtTexelRGBAlpha));
    for(i=0; i<texture_size; i++)
    {
      texture->array.rgb_alpha[i].red   = 
        (float)buf[i * BYTESPERPIXEL]     / 255.0;
      texture->array.rgb_alpha[i].green = 
        (float)buf[i * BYTESPERPIXEL + 1] / 255.0;
      texture->array.rgb_alpha[i].blue  = 
        (float)buf[i * BYTESPERPIXEL + 2] / 255.0;
      texture->array.rgb_alpha[i].alpha = 1.0;
    }
  break;

  case (PEXEnumTypeIndex)PEXExtTexelRGBInt8:
    texture->array.rgb8 =
      (PEXExtTexelRGB8 *)malloc(texture_size * sizeof(PEXExtTexelRGB8));
    for(i=0; i<texture_size; i++)
    {
      texture->array.rgb8[i].red   = buf[i * BYTESPERPIXEL];
      texture->array.rgb8[i].green = buf[i * BYTESPERPIXEL + 1];
      texture->array.rgb8[i].blue  = buf[i * BYTESPERPIXEL + 2];
    }
  break;

  case (PEXEnumTypeIndex)PEXExtTexelRGBAlphaInt8:
    texture->array.rgb_alpha8 = (PEXExtTexelRGBAlpha8 *)
      malloc(texture_size * sizeof(PEXExtTexelRGBAlpha8));
    for(i=0; i<texture_size; i++)
    {
      texture->array.rgb_alpha8[i].red   = buf[i * BYTESPERPIXEL];
      texture->array.rgb_alpha8[i].green = buf[i * BYTESPERPIXEL + 1];
      texture->array.rgb_alpha8[i].blue  = buf[i * BYTESPERPIXEL + 2];
      texture->array.rgb_alpha8[i].alpha = (unsigned char)255;
    }
  break;
  }
  free(buf);
}

struct Point *store_coord(struct Point *list, float x, float y, float z)
/*
 ******************************************************************************
 **    Description:
 **        This functions stores a coord into a linked list. 
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **
 **    Output parameters:
 **        None. 
 **
 **    Entry Conditions:
 **        1) Assume - that a texture filename exists. 
 **
 **    Exit Conditions:
 **        1) If texture file does not exist then Error() is called.
 **
 **    Error Conditions:
 **        See Exit Conditions.
 **
 **    Referenced by:
 **        Load_Object().
 **
 **    Procedure/Function dependencies:
 **        Load_Object().
 **
 ******************************************************************************
 */
{
  list->x = x;
  list->y = y;
  list->z = z;
  list->next = (struct Point *)malloc(sizeof(struct Point));
  list->next->next =  NULL;
  return(list->next);
}

 /* 
  * These are used for environment mapping to find the center of an object
  * Note: that environment mapping tests can have only one object
  */
  float minx, miny, minz, maxx, maxy, maxz;

void Load_Object(char *value, Input_Parameters *in, int obj)
/*
 ******************************************************************************
 **    Description:
 **        This functions reads a file containing object data into 
 **        a data structure that is used by PEX for rendering. 
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **
 **    Output parameters:
 **        None. 
 **
 **    Entry Conditions:
 **        1) Assume - that a texture filename exists. 
 **
 **    Exit Conditions:
 **        1) If texture file does not exist then Error() is called.
 **
 **    Error Conditions:
 **        See Exit Conditions.
 **
 **    Referenced by:
 **        Load_Object_Information().
 **
 **    Procedure/Function dependencies:
 **        Load_Object_Information().
 **
 ******************************************************************************
 */
{
 /* 
  * Used to keep track the total number of data read in
  */
  int POLYS = 0;
  int POINTS = 0;
  int NORMALS = 0;
  int COLORS = 0;
  int vertex_count;

 /* 
  * File pointer and input sting
  */
  FILE *fp;
  char incoming[1024];

 /*
  * Linked list pointers 
  */	
  struct Point *coord;
  struct Point *norm;
  struct Point *colr;
  struct Facet *facet;

  struct Point *list;
  struct Facet *flist;
  struct List *vert_list;
  struct List *norm_list;
  struct List *color_list;
  struct List *save;

 /*
  * Misc
  */
  int status;
  int vertex;
  int normal_ptr;
  int color_ptr;
  float x, y, z;
  int i, j;
  char errstr[ERRORSTRINGSIZE];

 /*
  * Used for storing the Facet data in PEX format
  */
  int fvertex_size;
  int fcolor_offset;
  int fnormal_offset;

 /*
  * Used for storing the vertices and vertex data in PEX format
  */
  int vertex_size;
  int point_offset;
  int color_offset;
  int normal_offset;
  int edge_offset;

 /*
  * Arrays 
  */
  struct Point *point;
  struct Point *normal;
  struct Point *color;
  float  *frgbvn;

  /* Open object file */
  fp = fopen(value, "r");
  if(fp == NULL)
  {
    sprintf(errstr, "Object file not found - '%s'",value);
    Error(errstr);
  }

  /* Get Width if it exists.  Note: this is only for Quad Mesh. */
  fscanf(fp,"%s",incoming);
  if(strcasecmp(incoming, "W") == 0)
  {
    fscanf(fp,"%d",&in->width[obj]);
    fscanf(fp,"%s",incoming);
  }

  /* Get Height if it exists.  Note: this is only for Quad Mesh. */
  if(strcasecmp(incoming, "H") == 0)
  {
    fscanf(fp,"%d",&in->height[obj]);
    fscanf(fp,"%s",incoming);
  }

  /* Allocate a head pointer for the vertex list */
  coord = (struct Point *)malloc(sizeof(struct Point));
  list = coord;

  /* Load vertex list */
  while(incoming[0] == 'v' && incoming[1] != 'n')
  {
    fgets(incoming, 80, fp);
    sscanf(incoming,"%f %f %f",&x, &y, &z);
    list = store_coord(list, x, y, z);
    POINTS++;
    fscanf(fp,"%s",incoming);
  }

  /* Allocate a head pointer for the normal list */
  norm = (struct Point *)malloc(sizeof(struct Point));
  list = norm;

  /* Load the normal list */
  while(incoming[0] == 'v' && incoming[1] == 'n')
  {
    fgets(incoming, 80, fp);
    sscanf(incoming,"%f %f %f",&x, &y, &z);
    list = store_coord(list, x, y, z);
    NORMALS++;
    fscanf(fp,"%s",incoming);
  }

  /* Allocate a head pointer for the color list */
  /* r will map to x, g will map to y, and b will map to x */
  /* This avoids more structures and functions */
  colr = (struct Point *)malloc(sizeof(struct Point));
  list = colr;

  /* Load the color list */
  while(incoming[0] == 'v' && incoming[1] == 'c')
  {
    fgets(incoming, 80, fp);
    sscanf(incoming,"%f %f %f",&x, &y, &z);
    list = store_coord(list, x, y, z);
    COLORS++;
    fscanf(fp,"%s",incoming);
  }

  /* Allocate a head pointer for the facet lists */
  facet = (struct Facet *)malloc(sizeof(struct Facet));
  facet->vert = NULL;
  facet->norm = NULL;
  facet->color = NULL;
  facet->next = NULL;
  flist = facet;

  /* Load facet information */
  while(incoming[0] == 'f' && incoming[1] == 'o')
  {
    vertex_count = 0;
    POLYS++;
    flist->vert = (struct List *)malloc(sizeof(struct List));
    flist->vert->next = NULL;
    flist->norm = (struct List *)malloc(sizeof(struct List));
    flist->norm->next = NULL;
    flist->color = (struct List *)malloc(sizeof(struct List));
    flist->color->next = NULL;
    vert_list = flist->vert;
    norm_list = flist->norm;
    color_list = flist->color;

    fscanf(fp, "%s", incoming);
    status = sscanf(incoming,"%d//%d", &vertex, &normal_ptr);
    if(status != 2)
      status = sscanf(incoming,"%d/%d/%d", &vertex, &color_ptr, &normal_ptr);
    while(status == 2 || status == 3)
    {
      vertex_count++;
      vert_list->index = vertex;
      vert_list->next = (struct List *)malloc(sizeof(struct List));
      vert_list = vert_list->next;
      vert_list->next = NULL;
      norm_list->index = normal_ptr;
      norm_list->next = (struct List *)malloc(sizeof(struct List));
      norm_list = norm_list->next;
      norm_list->next = NULL;

      if(status == 3)
      {
        color_list->index = color_ptr;
        color_list->next = (struct List *)malloc(sizeof(struct List));
        color_list = color_list->next;
        color_list->next = NULL;
      }

      incoming[0] = '\0';
      fscanf(fp, "%s", incoming);
      status = sscanf(incoming,"%d//%d", &vertex, &normal_ptr);
      if(status != 2)
        status = sscanf(incoming,"%d/%d/%d", &vertex, &color_ptr, &normal_ptr);
    }
    flist->vertex_count = vertex_count;
    flist->next = (struct Facet *)malloc(sizeof(struct Facet));
    flist = flist->next;
    flist->vert = NULL;
    flist->norm = NULL;
    flist->color = NULL;
    flist->next = NULL;
  }
  close(fp);

  /* Fill Sequential Point and Normal arrays. */
  /* I do this to make it easy to index in to 
     these arrays from the facet list */

  /* Fill points array and free up memory as you go */
  point = (struct Point *)malloc(POINTS*sizeof(struct Point));
  list = coord;

  /* This is for environment mapping */
  minx = maxx = list->x;
  miny = maxy = list->y;
  minz = maxz = list->z;

  for(i=0; i<POINTS; i++)
  {
    if(list->x > maxx)
      maxx = list->x;
    if(list->y > maxy)
      maxy = list->y;
    if(list->z > maxz)
      maxz = list->z;
    if(list->x < minx)
      minx = list->x;
    if(list->y < miny)
      miny = list->y;
    if(list->z < minz)
      minz = list->z;

    point[i].x = list->x;
    point[i].y = list->y;
    point[i].z = list->z;
    list = list->next;
    free(coord);
    coord = list;
  }

  /* Fill normal array and free up memory as you go */
  normal = (struct Point *)malloc(POINTS*sizeof(struct Point));
  list = norm;
  for(i=0; i<NORMALS; i++)
  {
    normal[i].x = list->x;
    normal[i].y = list->y;
    normal[i].z = list->z;

if (isnan(normal[i].x) || isnan(normal[i].y) || isnan(normal[i].z))
	fprintf(stderr, "You geek [%d]!  Bad data!  (%d)\n", i, obj);


    list = list->next;
    free(norm);
    norm = list;
  }

  /* Fill color array and free up memory as you go */
  color = (struct Point *)malloc(POINTS*sizeof(struct Point));
  list = colr;
  for(i=0; i<COLORS; i++)
  {
    color[i].x = list->x;
    color[i].y = list->y;
    color[i].z = list->z;
    list = list->next;
    free(colr);
    colr = list;
  }

  /* Calculate the offsets for all of the data */
  vertex_size = 3;
  point_offset = 0;
  if((in->vertex_att[obj] & PEXGAColor) == PEXGAColor)
  {
    color_offset = vertex_size;
    if((in->color_type[obj] & PEXHPColorTypeRGBA) == PEXHPColorTypeRGBA)
      vertex_size += 4;
    else
      vertex_size += 3;
  }
  else
    color_offset = 0;

  if((in->vertex_att[obj] & PEXGANormal) == PEXGANormal)
  {
    normal_offset = vertex_size;
    vertex_size += 3;
  }
  else
    normal_offset = 0;

  if((in->vertex_att[obj] & PEXGAEdges) == PEXGAEdges)
  {
    edge_offset = vertex_size;
    vertex_size += 1;
  }
  else
    edge_offset = 0;
  vertex_size += in->vertex_fp_data_size;

  /* Get Facet Data information */
  fvertex_size = 0;
  fcolor_offset = 0;
  if((in->facet_att[obj] & PEXGAColor) == PEXGAColor)
  {
    if((in->color_type[obj] & PEXHPColorTypeRGBA) == PEXHPColorTypeRGBA)
      fvertex_size += 4;
    else
      fvertex_size += 3;
  }

  if((in->facet_att[obj] & PEXGANormal) == PEXGANormal)
  {
    fnormal_offset = fvertex_size;
    fvertex_size += 3;
  }
  else
    fnormal_offset = 0;

 /*
  * Note:  In all cases I access the point and vertex list
  * with the facet pointer -1, since the file expects all 
  * facet indices to be one based (starting at 1), but the 
  * arrays are loaded 0 based.
  * BLS
  */

 
    /* Allocate a pointer for each facet */
    frgbvn = (float *)calloc(POLYS, fvertex_size * sizeof(float));

    /* Begining of the facet list */
    flist = facet;

    /* Traverse all of the facets */
    for (i=0; i<POLYS; i++)
    {
      j = 0;
      vert_list = flist->vert;
      norm_list = flist->norm;
      color_list = flist->color;

          if((in->facet_att[obj] & PEXGAColor) == PEXGAColor)
          {
            frgbvn[i*fvertex_size+0] =
              color[color_list->index-1].x;
            frgbvn[i*fvertex_size+1] =
              color[color_list->index-1].y;
            frgbvn[i*fvertex_size+2] =
              color[color_list->index-1].z;
            if((in->color_type[obj] & PEXHPColorTypeRGBA) == PEXHPColorTypeRGBA)
              frgbvn[i*fvertex_size+3] = .2;
          }
          if((in->facet_att[obj] & PEXGANormal) == PEXGANormal)
          {
            frgbvn[i*fvertex_size+fnormal_offset+0] = 
              normal[norm_list->index-1].x;
            frgbvn[i*fvertex_size+fnormal_offset+1] = 
              normal[norm_list->index-1].y;
            frgbvn[i*fvertex_size+fnormal_offset+2] = 
              normal[norm_list->index-1].z;
          }

      flist = flist->next;
    }

  switch(in->pex_coord_type[obj])
  {
  case FASWD: /* Fill Area Set With Data */
  {
    float  **rgbvn;
 
    /* Allocate space for the vertex list */
    in->vertex_list[obj] = (PEXExtListOfVertex *)
      malloc(sizeof(PEXExtListOfVertex)*POLYS);

    /* Allocate a pointer for each facet */
    rgbvn = (float **)calloc(POLYS, 4);

    /* Begining of the facet list */
    flist = facet;

    /* Traverse all of the facets */
    for (i=0; i<POLYS; i++)
    {
      /* Allocate vertex space for this facet */
      rgbvn[i] = (float *)
        calloc(flist->vertex_count, vertex_size * sizeof(float));

      vert_list = flist->vert;
      norm_list = flist->norm;
      color_list = flist->color;

      /* Traverse all of the vertices in this facet */
      for(j=0; j<flist->vertex_count; j++)
      {
        rgbvn[i][j*vertex_size+point_offset+0] = point[vert_list->index-1].x;
        rgbvn[i][j*vertex_size+point_offset+1] = point[vert_list->index-1].y;
        rgbvn[i][j*vertex_size+point_offset+2] = point[vert_list->index-1].z;
        if(color_offset)
        {
          rgbvn[i][j*vertex_size+color_offset+0] = 
            color[color_list->index-1].x;
          rgbvn[i][j*vertex_size+color_offset+1] = 
            color[color_list->index-1].y;
          rgbvn[i][j*vertex_size+color_offset+2] =
            color[color_list->index-1].z;
          if((in->color_type[obj] & PEXHPColorTypeRGBA) == PEXHPColorTypeRGBA)
            rgbvn[i][j*vertex_size+color_offset+3] = .2;
        }
        if(normal_offset)
        {
          rgbvn[i][j*vertex_size+normal_offset+0] = 
            normal[norm_list->index-1].x;
          rgbvn[i][j*vertex_size+normal_offset+1] = 
            normal[norm_list->index-1].y;
          rgbvn[i][j*vertex_size+normal_offset+2] = 
            normal[norm_list->index-1].z;
        }
        if(edge_offset)
        {
          rgbvn[i][j*vertex_size+edge_offset+0] = in->edge_flag;
        }

        vert_list = vert_list->next;
        norm_list = norm_list->next;
      }
      in->vertex_list[obj][i].count = flist->vertex_count;
      in->vertex_list[obj][i].vertices.with_fp_data = (PEXPointer)rgbvn[i];
      flist = flist->next;
    }
    in->facet_data = (PEXFacetData *)frgbvn;
  }
  break; /* End Case FASWD */

  case SOFAS: /* Set Of Fill Area Sets */
  {
    float *rgbvn;
    unsigned short **indices;
    PEXListOfUShort *contours;

    /* Allocate space for the vertex list */
    in->connectivity[obj] = (PEXConnectivityData *)
      malloc(POLYS * sizeof(PEXConnectivityData));
    contours = (PEXListOfUShort *)malloc(POLYS* sizeof(PEXListOfUShort));

    in->facet_array.index = NULL;

    /* Allocate a pointer for each vertex */
    rgbvn = (float *)calloc(POINTS, vertex_size * sizeof(float));

    /* Store all vertex information directly */
    for(i=0; i<POINTS; i++)
    {
      rgbvn[i*vertex_size+point_offset+0] = point[i].x;
      rgbvn[i*vertex_size+point_offset+1] = point[i].y;
      rgbvn[i*vertex_size+point_offset+2] = point[i].z;
    }
      
    /* Load the contour and connectivity data */
    indices = (unsigned short **)calloc(POLYS, sizeof(unsigned short *));
    in->index_count[obj] = 0;
    flist = facet;
    for (i=0; i<POLYS; i++)
    {
      indices[i] = (unsigned short *)
        malloc(flist->vertex_count * sizeof(unsigned short));
      vert_list = flist->vert;
      norm_list = flist->norm;
      color_list = flist->color;
      for(j=0; j<flist->vertex_count; j++)
      {
        indices[i][j] = vert_list->index-1;

        /* While traversing the facet list store normal and color information 
           with the corresponding point */
        if(color_offset)
        {
          rgbvn[(vert_list->index-1)*vertex_size+color_offset+0] = 
            color[color_list->index-1].x;
          rgbvn[(vert_list->index-1)*vertex_size+color_offset+1] = 
            color[color_list->index-1].y;
          rgbvn[(vert_list->index-1)*vertex_size+color_offset+2] = 
            color[color_list->index-1].z;
          if((in->color_type[obj] & PEXHPColorTypeRGBA) == PEXHPColorTypeRGBA)
            rgbvn[(vert_list->index-1)*vertex_size+color_offset+3] = .5;

          color_list = color_list->next;
        }
        if(normal_offset)
        {
          rgbvn[(vert_list->index-1)*vertex_size+normal_offset+0] = 
            normal[norm_list->index-1].x;
          rgbvn[(vert_list->index-1)*vertex_size+normal_offset+1] = 
            normal[norm_list->index-1].y;
          rgbvn[(vert_list->index-1)*vertex_size+normal_offset+2] = 
            normal[norm_list->index-1].z;

          norm_list = norm_list->next;
        }
        vert_list = vert_list->next;
      }
      contours[i].count = flist->vertex_count;
      in->index_count[obj] += flist->vertex_count;
      contours[i].shorts = indices[i];
      in->connectivity[obj][i].count = 1;
      in->connectivity[obj][i].lists = &contours[i];
      flist = flist->next;
    }
    in->vertex_array[obj].with_fp_data = (PEXPointer)rgbvn;
    in->facet_array.rgb_normal = (PEXColorRGBNormal *)frgbvn;
  }
  break; /* End Case SOFAS */

  case TS: /* Tri Strip */
  case QM: /* Quad Mesh */
  {
    float *rgbvn;

    /* Allocate a pointer for each vertex */
    rgbvn = (float *)calloc(POINTS, vertex_size * sizeof(float));

    for(i=0; i<POINTS; i++)
    {
      rgbvn[i*vertex_size+point_offset+0] = point[i].x;
      rgbvn[i*vertex_size+point_offset+1] = point[i].y;
      rgbvn[i*vertex_size+point_offset+2] = point[i].z;
      if(color_offset)
      {
        rgbvn[i*vertex_size+color_offset+0] = color[i].x;
        rgbvn[i*vertex_size+color_offset+1] = color[i].y;
        rgbvn[i*vertex_size+color_offset+2] = color[i].z;
      }
      if(normal_offset)
      {
        rgbvn[i*vertex_size+normal_offset+0] = normal[i].x;
        rgbvn[i*vertex_size+normal_offset+1] = normal[i].y;
        rgbvn[i*vertex_size+normal_offset+2] = normal[i].z;
      }
    }
    in->vertex_array[obj].with_fp_data = (PEXPointer)rgbvn;
    in->facet_array.rgb_normal = (PEXColorRGBNormal *)frgbvn;
  }
  break; /* End Case QM */
  }
  in->POLYS[obj] = POLYS;
  in->count[obj] = POLYS;
  in->POINTS[obj] = POINTS;

  /* Free everything */
  free(point);
  free(normal);
  free(color);
  
  flist = facet;
  while(flist->next != NULL)
  {
    vert_list = flist->vert;
    save = vert_list;
    while(vert_list != NULL)
    {
      vert_list = vert_list->next;
      free(save);
      save = vert_list;
    }
      
    norm_list = flist->norm;
    save = norm_list;
    while(norm_list != NULL)
    {
      norm_list = norm_list->next;
      free(save);
      save = norm_list;
    }
      
    color_list = flist->color;
    save = color_list;
    while(color_list != NULL)
    {
      color_list = color_list->next;
      free(save);
      save = color_list;
    }
      
    flist = flist->next;
    free(facet);
    facet = flist;
  } /* End Free enverything */
}

Load_Alpha_Data(PEXExtTexelArray *texture,
                PEXExtTexelArray alpha,
                int channel_hint,
                PEXEnumTypeIndex texel_type)
/*
 ******************************************************************************
 **    Description:
 **        This function copies alpha information in to a texture map.
 **
 **    Author:
 **       Brad Saunders                Jan. 06, 1994
 **
 **    Input parameters:
 **       PEXExtTexelArray *texture - pointer to texture.
 **       PEXExtTexelArray alpha - alpha map.
 **       int channel_hint - gives channel that has alpha information.
 **       PEXEnumTypeIndex texel_type - texel type.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        None.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Load_Texture_Information().
 **
 **    Procedure/Function dependencies:
 **        Load_Texture_Information().
 **
 ******************************************************************************
 */
{
  char errstr[ERRORSTRINGSIZE];
  long alpha_size;
  int i;

  alpha_size = alpha.dimension.t0 * alpha.dimension.t1;

  switch(texel_type)
  {
  case (PEXEnumTypeIndex)PEXExtTexelRGBAlphaFloat:
    for(i=0; i<alpha_size; i++)
    {
      switch(channel_hint)
      {
      case 1: /* RED */
        texture->array.rgb_alpha[i].alpha = 
          alpha.array.rgb_alpha[i].red;
       break;

      case 2: /* GREEN */
        texture->array.rgb_alpha[i].alpha = 
          alpha.array.rgb_alpha[i].green;
       break;

      case 3: /* BLUE */
        texture->array.rgb_alpha[i].alpha = 
          alpha.array.rgb_alpha[i].blue;
       break;

      default:
        sprintf(errstr, "'%d' is an invalid alpha hint", channel_hint);
      break;
      }
    }
  break;

  case (PEXEnumTypeIndex)PEXExtTexelRGBAlphaInt8:
    for(i=0; i<alpha_size; i++)
    {
      switch(channel_hint)
      {
      case 1: /* RED */
        texture->array.rgb_alpha8[i].alpha = alpha.array.rgb_alpha8[i].red;
       break;

      case 2: /* GREEN */
        texture->array.rgb_alpha8[i].alpha = alpha.array.rgb_alpha8[i].green;
       break;

      case 3: /* BLUE */
        texture->array.rgb_alpha8[i].alpha = alpha.array.rgb_alpha8[i].blue;
       break;

      default:
        sprintf(errstr, "'%d' is an invalid alpha hint", channel_hint);
      break;
      }
    }
  break;

  default:
    sprintf(errstr, "No alpha data loaded");
    Warning(errstr);
  break;
  }
}

void Load_Texture_Information(char **file, Input_Parameters *in)
/*
 ******************************************************************************
 **    Description:
 **        This function loads all of the texture information from
 **        the parameter file.
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char *file - a pointer to the input file.
 **        Input_Parameters in - stores the information being read.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  int i, x, y;
  char value[INPUTSIZE];
  char errstr[ERRORSTRINGSIZE];
  int alpha_present;
  
 /*
  * Allocate space to hold texture information times the total 
  * number of textures.
  */
  in->domain = (int *)malloc(sizeof(int) * in->TEXTURES);
  in->domain_data = 
    (PEXExtTMDomainData *)malloc(sizeof(PEXExtTMDomainData) * in->TEXTURES);
  in->call_create_filtered_tm = (int *)malloc(sizeof(int) * in->TEXTURES);
  in->call_create_filtered_tm_from_window = 
    (int *)malloc(sizeof(int) * in->TEXTURES);
  in->luminance_selector = (unsigned int *)malloc(sizeof(int) * in->TEXTURES);
  in->alpha_selector = (unsigned int *)malloc(sizeof(int) * in->TEXTURES);
  in->window_depth = (int *)malloc(sizeof(int) * in->TEXTURES);
  in->base_map = 
    (PEXExtTexelArray *)malloc(sizeof(PEXExtTexelArray) * in->TEXTURES);
  in->alpha_map = 
    (PEXExtTexelArray *)malloc(sizeof(PEXExtTexelArray) * in->TEXTURES);
  in->alpha_channel_hint = (int *)malloc(sizeof(int) * in->TEXTURES);
  in->texel_array = 
    (PEXExtTexelArray **)calloc(in->TEXTURES, sizeof(PEXExtTexelArray *));
  for(i=0; i<in->TEXTURES; i++)
    in->texel_array[i] = 
      (PEXExtTexelArray *)malloc(sizeof(PEXExtTexelArray *));
  in->parameterization = (int *)malloc(sizeof(int) * in->TEXTURES);
  in->param_data = 
    (PEXExtTMParameterizationData *)
    malloc(sizeof(PEXExtTMParameterizationData) * in->TEXTURES);
  in->rendering_order = (int *)malloc(sizeof(int) * in->TEXTURES);

  /* For each texture load appropriate information */
  for(i=0; i<in->TEXTURES; i++)
  {
    /* Get the texture number  */
    Get_Keyword_Value(file, "Texture_number", value);
    in->texture_indx = atoi(value);

    /* Get texture mapping domain */
    Get_Keyword_Value(file, "Texture_mapping_domain", value);
    in->domain[i] = Convert_Constant(value);

    /* Get texture mapping type */
    Get_Keyword_Value(file, "Texture_mapping_type", value);
    in->domain_data[i].data.color.tm_type = Convert_Enum(value);

    /* Get texel type */
    Get_Keyword_Value(file, "Texel_type", value);
    in->domain_data[i].data.color.texel_type = Convert_Enum(value);

    /* Get number of map levels */
    Get_Keyword_Value(file, "Number_of_map_levels", value);
    in->domain_data[i].data.color.num_levels = (unsigned short)atoi(value);

    /* Get texture map filename */
    Get_Keyword_Value(file, "Texture_filename", value);
    if(strcasecmp(value, "NULL") != 0)
      Load_Map(&in->base_map[i], 
               in->domain_data[i].data.color.texel_type, 
               value);

    /* Get Alpha map filename */
    Get_Keyword_Value(file, "Alpha_filename", value);
    if((in->domain_data[i].data.color.texel_type == 
       (PEXEnumTypeIndex)PEXExtTexelRGBAlphaFloat ||
       in->domain_data[i].data.color.texel_type == 
       (PEXEnumTypeIndex)PEXExtTexelRGBAlphaInt8) && 
       strcasecmp(value, "NULL") != 0)
    {
      alpha_present = True;
      Load_Map(&in->alpha_map[i], 
               in->domain_data[i].data.color.texel_type, 
               value);
      if(in->base_map[i].dimension.t0 != in->alpha_map[i].dimension.t0 ||
         in->base_map[i].dimension.t1 != in->alpha_map[i].dimension.t1)
      {
        sprintf(errstr, 
          "Inconsistant texture map and alpha map dimensions\n"  
          "for texture number %d", in->texture_indx);
        Warning(errstr);
      }
    }
    else
      alpha_present = False;

    /* Get Alpha map hint */
    Get_Keyword_Value(file, "Alpha_channel_hint", value);
    in->alpha_channel_hint[i] = atoi(value);

    /* Get do filtered TM */
    Get_Keyword_Value(file, "Do_CreateFilteredTM", value);
    in->call_create_filtered_tm[i] = 
      Convert_Flags(value, "Do_CreateFilteredTM");

    /* Get do filtered TM From Window */
    Get_Keyword_Value(file, "Do_CreateFilteredTMFromWindow", value);
    in->call_create_filtered_tm_from_window[i] = 
      Convert_Flags(value, "Do_CreateFilteredTMFromWindow");

    /* Get Luminance Channel Selector */
    Get_Keyword_Value(file, "Luminance_channel_selector", value);
    in->luminance_selector[i] = (unsigned int)Convert_Constant(value);

    /* Get Alpha Channel Selector */
    Get_Keyword_Value(file, "Alpha_channel_selector", value);
    in->alpha_selector[i] = (unsigned int)Convert_Constant(value);

    /* Get Window Depth */
    Get_Keyword_Value(file, "Window_depth", value);
    in->window_depth[i] = atoi(value);

    if(in->call_create_filtered_tm_from_window[i] == False && alpha_present)
      Load_Alpha_Data(&in->base_map[i], 
                      in->alpha_map[i],
                      in->alpha_channel_hint[i],
                      in->domain_data[i].data.color.texel_type);

    /* For PEXExtCreateTMDescription */

    /* Get Parameterization */
    Get_Keyword_Value(file, "Parameterization", value);
    in->parameterization[i] = Convert_Constant(value);

    /* Get Reflection matrix */
    Get_Value(file, value);
    if(strcmp("Reflection_matrix", value) != 0)
      Error("Reflection_matrix");
    for(x=0; x<4; x++)
      for(y=0; y<4; y++)
      {
        Get_Value(file, value);
        in->param_data[i].data.reflection.matrix[x][y]
          = atof(value);
      }
    /* Get plane equations */
    Get_Value(file, value);
    if(strcmp("P0", value) != 0)
      Error("P0 not found in TM Description");
    for(x=0; x<4; x++)
    {
      Get_Value(file, value);
      if(in->parameterization[i] == PEXExtTMParamLinearVRC)
        in->param_data[i].data.linear.p0[x] = atof(value);
    }

    Get_Value(file, value);
    if(strcmp("P1", value) != 0)
      Error("P1 not found in TM Description");
    for(x=0; x<4; x++)
    {
      Get_Value(file, value);
      if(in->parameterization[i] == PEXExtTMParamLinearVRC)
        in->param_data[i].data.linear.p1[x] = atof(value);
    }

    /* Get Rendering Order */
    Get_Keyword_Value(file, "TM_rendering_order", value);
    in->rendering_order[i] = Convert_Constant(value);

  } /* end get texture info */
}

void Load_Object_Information(char **file, Input_Parameters *in)
/*
 ******************************************************************************
 **    Description:
 **        This function loads all of the object information from
 **        the parameter file.
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char *file - a pointer to the input file.
 **        Input_Parameters in - stores the information being read.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  int x, y;
  char value[INPUTSIZE];

  /* Get Projection Method */
  Get_Keyword_Value(file, "Projection_method", value);
  in->coord_data.projection = Convert_Enum(value);

  /* Get Reflection matrix */
  Get_Value(file, value);
  if(strcmp("Reflection_matrix", value) != 0)
    Error("Reflection_matrix not found in Object Coord Data");
  for(x=0; x<4; x++)
    for(y=0; y<4; y++)
    {
      Get_Value(file, value);
      in->coord_data.param_data.data.reflection.matrix[x][y]
        = atof(value);
    }

  /* Get plane equations */
  Get_Value(file, value);
  if(strcmp("P0", value) != 0)
    Error("P0 not found in Object Coord Data");
  for(x=0; x<4; x++)
  {
    Get_Value(file, value);
    if(in->coord_data.projection == PEXExtTMProjectionLinearWC)
      in->coord_data.param_data.data.linear.p0[x] = atof(value);
  }

  Get_Value(file, value);
  if(strcmp("P1", value) != 0)
    Error("P1 not found in Object Coord Data");
  for(x=0; x<4; x++)
  {
    Get_Value(file, value);
    if(in->coord_data.projection == PEXExtTMProjectionLinearWC)
      in->coord_data.param_data.data.linear.p1[x] = atof(value);
  }

  /* Get coord source */
  Get_Keyword_Value(file, "Coord_source", value);
  in->coord_data.coord_source = Convert_Enum(value);

  /* Get fp data index */
  Get_Keyword_Value(file, "Fp_data_index", value);
  in->coord_data.fp_data_index = (unsigned short)atoi(value);

  /* Get transform matrix */
  Get_Value(file, value);
  if(strcmp("Model_transform_matrix", value) != 0)
    Error("Model_transform_matrix not found in Object Coord Data");
  for(x=0; x<4; x++)
    for(y=0; y<4; y++)
    {
      Get_Value(file, value);
      in->coord_data.mc_transform[x][y] = atof(value);
    }

  /* Get vertex fp data size */
  Get_Keyword_Value(file, "Vertex_fp_data_size", value);
  in->vertex_fp_data_size = (unsigned int)atoi(value);

/* Begin support of MULTIPLE objects */
in->vertex_att = (unsigned int *)malloc(MAX_OBJS*sizeof(unsigned int));
in->facet_att = (unsigned int *)malloc(MAX_OBJS*sizeof(unsigned int));
in->color_type = (int *)malloc(MAX_OBJS*sizeof(int));
in->pex_coord_type = (int *)malloc(MAX_OBJS*sizeof(int));
in->index_count = (unsigned int *)malloc(MAX_OBJS*sizeof(unsigned int));
in->POINTS = (unsigned int *)malloc(MAX_OBJS*sizeof(unsigned int));
in->POLYS = (unsigned int *)malloc(MAX_OBJS*sizeof(unsigned int));
in->count = (unsigned int *)malloc(MAX_OBJS*sizeof(unsigned int));
in->width = (unsigned int *)malloc(MAX_OBJS*sizeof(unsigned int));
in->height = (unsigned int *)malloc(MAX_OBJS*sizeof(unsigned int));
in->vertex_list = 
  (PEXExtListOfVertex **)calloc(MAX_OBJS, sizeof(PEXExtListOfVertex *));
in->connectivity = 
  (PEXConnectivityData **)calloc(MAX_OBJS, sizeof(PEXConnectivityData *));
in->vertex_array = 
  (PEXExtArrayOfVertex *)calloc(MAX_OBJS, sizeof(PEXExtArrayOfVertex *));

in->obj = 0;
Get_Value(file,value);
while(strcmp(value, "Vertex_attributes") == 0)
{
if(in->obj > MAX_OBJS)
  Error("Maximum number of objects exceeded\n");

  /* These will handle more than on flag */
  /* Get vertex attributes */
  Get_Value(file, value);
  in->vertex_att[in->obj] = Convert_Constant(value);
  Get_Value(file, value);
  while(strcmp(value, "|") == 0)
  {
    Get_Value(file, value);
    in->vertex_att[in->obj] |= Convert_Constant(value);
    Get_Value(file, value);
  }

  /* Get facet attributes*/
  if(strcmp(value, "Facet_attributes") != 0)
    Error("Facet_attributes not found in Object Coord Data");
  Get_Value(file, value);
  in->facet_att[in->obj] = Convert_Constant(value);
  Get_Value(file, value);
  while(strcmp(value, "|") == 0)
  {
    Get_Value(file, value);
    in->facet_att[in->obj] |= Convert_Constant(value);
    Get_Value(file, value);
  }

  /* Get color type */
  if(strcmp(value, "Color_type") != 0)
    Error("Color_type not found in Object Coord Data");
  Get_Value(file, value);
  in->color_type[in->obj] = Convert_Constant(value);

  /* Get object primitive type */
  Get_Keyword_Value(file, "PEX_coord_type", value);
  if(strcmp(value, "FASWD") == 0)
    in->pex_coord_type[in->obj] = FASWD;
  else
     if(strcmp(value, "SOFAS") == 0)
       in->pex_coord_type[in->obj] = SOFAS;
     else
       if(strcmp(value, "TS") == 0)
         in->pex_coord_type[in->obj] = TS;
       else
         if(strcmp(value, "QM") == 0)
           in->pex_coord_type[in->obj] = QM;
         else
           Error("No correct PEX Coord Type");
 

  /* Get object filename */
  Get_Keyword_Value(file, "Object_filename", value);
  Load_Object(value, in, in->obj);
Get_Value(file,value);
in->obj++;
}
*file -=18;
}
  
void Load_OC_Information(char **file, Input_Parameters *in)
/*
 ******************************************************************************
 **    Description:
 **        This function loads all of the OC information from
 **        the parameter file.
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char *file - a pointer to the input file.
 **        Input_Parameters in - stores the information being read.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  int i;
  char value[INPUTSIZE];

  /* Get Optimization Hint */
  Get_Keyword_Value(file, "Optimization_hint", value);
  in->opt_hint = Convert_Constant(value);

  /* Get number of members in priorities */
  Get_Keyword_Value(file, "Count", value);
  in->opt_count = (unsigned)atoi(value);
  in->priorities = (PEXTableIndex *)malloc(in->opt_count * 
                                          sizeof(PEXTableIndex));
  Get_Value(file, value);
  if(strcmp(value, "Priority_list") != 0)
    Error("Priority_list not found in PEXExtSetTMResourceHints");
  for(i=0; i<in->opt_count; i++)
  {
    Get_Value(file, value);
    in->priorities[i] = (PEXTableIndex)atoi(value);
  }

  /* Get number of Front Face textures */
  Get_Keyword_Value(file, "Number_of_textures_to_activate", value);
  in->number_FF_textures = (unsigned)atoi(value);
  in->FF_texture_indices = (PEXTableIndex *)malloc(in->number_FF_textures * 
                                                  sizeof(PEXTableIndex));
  Get_Value(file, value);
  if(strcmp(value, "FF_active_textures_list") != 0)
    Error("FF_active_textures_list not found");
  for(i=0; i<in->number_FF_textures; i++)
  {
    Get_Value(file, value);
    in->FF_texture_indices[i] = (PEXTableIndex)atoi(value);
  }

  /* Get number of Back Face textures */
  Get_Keyword_Value(file, "Number_of_textures_to_activate", value);
  in->number_BF_textures = (unsigned)atoi(value);
  in->BF_texture_indices = (PEXTableIndex *)malloc(in->number_BF_textures * 
                                                  sizeof(PEXTableIndex));
  Get_Value(file, value);
  if(strcmp(value, "BF_active_textures_list") != 0)
    Error("BF_active_textures_list not found");
  for(i=0; i<in->number_BF_textures; i++)
  {
    Get_Value(file, value);
    in->BF_texture_indices[i] = (PEXTableIndex)atoi(value);
  }

  /* Get perspective correction method */
  Get_Keyword_Value(file, "Method", value);
  in->method = Convert_Constant(value);

  /* Get sampling frequency */
  Get_Keyword_Value(file, "Frequency", value);
  in->frequency = Convert_Constant(value);
}

void Load_Binding_LUT_Information(char **file, Input_Parameters *in)
/*
 ******************************************************************************
 **    Description:
 **        This function loads all of the Binding LUT information from
 **        the parameter file.
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char *file - a pointer to the input file.
 **        Input_Parameters in - stores the information being read.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  int i;
  char value[INPUTSIZE];

  /* Get number of binding LUT entries */
  Get_Keyword_Value(file, "Number_of_binding_LUT_entries", value);
  in->num_binding = atoi(value);
  in->binding_info = (int *)malloc(in->num_binding * sizeof(int));
  in->binding_indx = (int *)malloc(in->num_binding * sizeof(int));
  in->binding = (PEXExtTMBindingEntry *)malloc(in->num_binding * 
               	sizeof(PEXExtTMBindingEntry));

  for(i=0; i<in->num_binding; i++)
  {
    /* Get binding LUT index */
    Get_Keyword_Value(file, "Binding_LUT_index_number", value);
    in->binding_indx[i] = atoi(value);

    /* Get texture map number */
    Get_Keyword_Value(file, "Texture_map_number", value);
    in->binding_info[i] = atoi(value);

    /* Get Coord Source index */
    Get_Keyword_Value(file, "Coordsource_index", value);
    in->binding[i].coord_source_index = (PEXTableIndex)atoi(value);

    /* Get composition index */
    Get_Keyword_Value(file, "Composition_index", value);
    in->binding[i].composition_index = (PEXTableIndex)atoi(value);

    /* Get sampling index */
    Get_Keyword_Value(file, "Sampling_index", value);
    in->binding[i].sampling_index = (PEXTableIndex)atoi(value);
  }
}

void Load_CoordSource_LUT_Information(char **file, Input_Parameters *in)
/*
 ******************************************************************************
 **    Description:
 **        This function loads all of the Coord Source LUT information 
 **        from the parameter file.
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char *file - a pointer to the input file.
 **        Input_Parameters in - stores the information being read.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  int i, x, y;
  char value[INPUTSIZE];

  /* Get number of Coord Source LUT entries */
  Get_Keyword_Value(file, "Number_of_coordsource_LUT_entries", value);
  in->num_coordsource = atoi(value);
  in->coordsource = 
    (PEXExtTMCoordSourceEntry *)malloc(in->num_coordsource * 
                                         sizeof(PEXExtTMCoordSourceEntry));
  in->coordsource_indx = (int *) malloc(in->num_coordsource *
                                            sizeof(int));

  for(i=0; i<in->num_coordsource; i++)
  {
    /* Get Coord Source LUT index */
    Get_Keyword_Value(file, "Coordsource_LUT_index_number", value);
    in->coordsource_indx[i] = atoi(value);

    /* Get TM source */
    Get_Keyword_Value(file, "Tm_source", value);
    in->coordsource[i].tm_source = Convert_Enum(value);

    /* Get TM coord index */
    Get_Keyword_Value(file, "Fp_data_index", value);
    in->coordsource[i].fp_data_index = (unsigned short)atoi(value);

    /* Get orientation matrix */
    Get_Value(file, value);
    if(strcmp("Orientation_matrix", value) != 0)
      Error("Orientation_matrix not found in Coord Source LUT");
    for(x=0; x<4; x++)
      for(y=0; y<4; y++)
      {
        Get_Value(file, value);
        in->coordsource[i].orientation[x][y] = atof(value);
      }
  }
}


void Load_Composition_LUT_Information(char **file, Input_Parameters *in)
/*
 ******************************************************************************
 **    Description:
 **        This function loads all of the Composition LUT information
 **        from the parameter file.
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char *file - a pointer to the input file.
 **        Input_Parameters in - stores the information being read.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  int i;
  char value[INPUTSIZE];

  /* Get number of composition LUT entries */
  Get_Keyword_Value(file, "Number_of_composition_LUT_entries", value);
  in->num_composition = atoi(value);
  in->composition = (PEXExtTMCompositionEntry *)malloc(in->num_composition * 
                   sizeof(PEXExtTMCompositionEntry));
  in->composition_indx = (int *) malloc(in->num_composition * sizeof(int));

  for(i=0; i<in->num_composition; i++)
  {
    /* Get composition LUT index */
    Get_Keyword_Value(file, "Composition_LUT_index_number", value);
    in->composition_indx[i] = atoi(value);

    /* Get operaztion */
    Get_Keyword_Value(file, "Method", value);
    in->composition[i].method = Convert_Enum(value);
  }
}

void Load_Sampling_LUT_Information(char **file, Input_Parameters *in)
/*
 ******************************************************************************
 **    Description:
 **        This function loads all of the Sampling LUT information
 **        from the parameter file.
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char *file - a pointer to the input file.
 **        Input_Parameters in - stores the information being read.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  int i;
  char value[INPUTSIZE];

  /* Get number of sampling LUT entries */
  Get_Keyword_Value(file, "Number_of_sampling_LUT_entries", value);
  in->num_sampling = atoi(value);
  in->sampling = (PEXExtTMSamplingEntry *)malloc(in->num_sampling * 
                                      sizeof(PEXExtTMSamplingEntry));
  in->sampling_indx = (int *) malloc(in->num_sampling * sizeof(int));

  for(i=0; i<in->num_sampling; i++)
  {
    /* Get sampling LUT index */
    Get_Keyword_Value(file, "Sampling_LUT_index_number", value);
    in->sampling_indx[i] = atoi(value);

    /* Get minification method */
    Get_Keyword_Value(file, "Minification_method", value);
    in->sampling[i].minification_method = Convert_Enum(value);

    /* Get magnification method */
    Get_Keyword_Value(file, "Magnification_method", value);
    in->sampling[i].magnification_method = Convert_Enum(value);

    /* Get t0 boundry condition */
    Get_Keyword_Value(file, "T0_boundry_condition", value);
    in->sampling[i].t0_boundary_condition = Convert_Enum(value);

    /* Get t1 boundry condition */
    Get_Keyword_Value(file, "T1_boundry_condition", value);
    in->sampling[i].t1_boundary_condition = Convert_Enum(value);

    /* Get t2 boundry condition */
    Get_Keyword_Value(file, "T2_boundry_condition", value);
    in->sampling[i].t2_boundary_condition = Convert_Enum(value);

    /* Get boundry clamp color source */
    Get_Keyword_Value(file, "Boundry_clamp_color_source", value);
    in->sampling[i].boundary_clamp_color_source = Convert_Enum(value);

    /* Get Color Specifier */
    Get_Color_Specifier(file, &in->sampling[i].clamp_color, "Sampling LUT");

    /* Get depth sampling bias hint */
    Get_Keyword_Value(file, "Depth_sampling_bias_hint", value);
    in->sampling[i].depth_sampling_bias_hint = atof(value);

    /* Get t0 frequency hint */
    Get_Keyword_Value(file, "T0_frequency_hint", value);
    in->sampling[i].t0_frequency_hint = atof(value);

    /* Get t1 frequency hint */
    Get_Keyword_Value(file, "T1_frequency_hint", value);
    in->sampling[i].t1_frequency_hint = atof(value);

    /* Get t2 frequency hint */
    Get_Keyword_Value(file, "T2_frequency_hint", value);
    in->sampling[i].t2_frequency_hint = atof(value);
 }
}

unsigned short Convert_Clip_Flags(char *value)
/*
 ******************************************************************************
 **    Description:
 **        This functionconverts clip flag strings to values. 
 **
 **    Author:
 **       Brad Saunders                Dec. 17, 1993
 **
 **    Input parameters:
 **        char *value - a pointer to a clip flag string.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Load_Viewport_Information().
 **
 **    Procedure/Function dependencies:
 **        None.
 **
 ******************************************************************************
 */
{
  char errstr[ERRORSTRINGSIZE];

  if(strcmp(value, "PEXClipXY") == 0)
    return(PEXClipXY);
  else
    if(strcmp(value, "PEXClipBack") == 0)
      return(PEXClipBack);
    else
      if(strcmp(value, "PEXClipFront") == 0)
        return(PEXClipFront);
      else
        if(strcmp(value, "PEXClippingAll") == 0)
          return(PEXClippingAll);
        else
        {
          sprintf(errstr,"Clipping not found - '%s'",value);
          Error(errstr);
        }
}


void Load_Viewport_Information(char **file, Input_Parameters *in)
/*
 ******************************************************************************
 **    Description:
 **        This function loads all of the Viewport information
 **        from the parameter file.
 **
 **    Author:
 **       Brad Saunders                Dec. 16, 1993
 **
 **    Input parameters:
 **        char *file - a pointer to the input file.
 **        Input_Parameters in - stores the information being read.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  char value[INPUTSIZE];

  Get_Keyword_Value(file, "Do_perspective_projection", value);
  in->do_perspective = Convert_Flags(value, "Do_perspective_projection");

  /* This will handle more than on flag */
  /* Get View Clip Flag */
  Get_Keyword_Value(file, "View_clip_flag", value);
  in->view.clip_flags = Convert_Clip_Flags(value);
  Get_Value(file, value);
  while(strcmp(value, "|") == 0)
  {
    Get_Value(file, value);
    in->view.clip_flags |= Convert_Clip_Flags(value);
    Get_Value(file, value);
  }

  /* Get View Reference Point */
  if(strcmp(value, "View_reference_point") != 0)
    Error("Keyword - 'View_reference_point' not found");
  Get_Value(file, value);
  in->view_ref_pt.x = atof(value);
  Get_Value(file, value);
  in->view_ref_pt.y = atof(value);
  Get_Value(file, value);
  in->view_ref_pt.z = atof(value);

  /* Get View Plane Normal */
  Get_Keyword_Value(file, "View_plane_normal", value);
  in->view_plane_normal.x = atof(value);
  Get_Value(file, value);
  in->view_plane_normal.y = atof(value);
  Get_Value(file, value);
  in->view_plane_normal.z = atof(value);

  /* Get View Up Vector */
  Get_Keyword_Value(file, "View_up_vec", value);
  in->view_up_vec.x = atof(value);
  Get_Value(file, value);
  in->view_up_vec.y = atof(value);
  Get_Value(file, value);
  in->view_up_vec.z = atof(value);

  /* Get Projection Reference Point */
  Get_Keyword_Value(file, "Projection_reference_point", value);
  in->prp.x = atof(value);
  Get_Value(file, value);
  in->prp.y = atof(value);
  Get_Value(file, value);
  in->prp.z = atof(value);

  /* Get View Frame 0 */
  Get_Keyword_Value(file, "View_frame_0", value);
  in->window[0].x = atof(value);
  Get_Value(file, value);
  in->window[0].y = atof(value);

  /* Get View Frame 1 */
  Get_Keyword_Value(file, "View_frame_1", value);
  in->window[1].x = atof(value);
  Get_Value(file, value);
  in->window[1].y = atof(value);

  /* Get Front Plane */
  Get_Keyword_Value(file, "Front_plane", value);
  in->front_plane = atof(value);

  /* Get View Plane */
  Get_Keyword_Value(file, "View_plane", value);
  in->view_plane = atof(value);

  /* Get Back Plane */
  Get_Keyword_Value(file, "Back_plane", value);
  in->back_plane = atof(value);

  /* Get Viewport Min xyz */
  Get_Keyword_Value(file, "Viewport_min_xyz", value);
  in->viewport.min.x = atof(value);
  Get_Value(file, value);
  in->viewport.min.y = atof(value);
  Get_Value(file, value);
  in->viewport.min.z = atof(value);

  /* Get Viewport Max xyz */
  Get_Keyword_Value(file, "Viewport_max_xyz", value);
  in->viewport.max.x = atof(value);
  Get_Value(file, value);
  in->viewport.max.y = atof(value);
  Get_Value(file, value);
  in->viewport.max.z = atof(value);

  in->view.clip_limits = in->viewport;
}

void Get_Color_Specifier(char **file, PEXColorSpecifier *color, char *keyword)
/*
 ******************************************************************************
 **    Description:
 **        This function loads all of the Viewport information
 **        from the parameter file.
 **
 **    Author:
 **       Brad Saunders                Dec. 16, 1993
 **
 **    Input parameters:
 **        char *file - a pointer to the input file.
 **        Input_Parameters in - stores the information being read.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Load_Lighting_Information().
 **        Load_Sampling_LUT_Information().
 **
 **    Procedure/Function dependencies:
 **        None.
 **
 ******************************************************************************
 */
{
  char value[INPUTSIZE];
  int set = 0;
  char errstr[ERRORSTRINGSIZE];

  /* Get Do Color Indexed */
  Get_Keyword_Value(file, "Do_color_indexed", value);
  if(strcasecmp(value, "True") == 0)
  {
    color->type = PEXColorTypeIndexed;
    set = 1;
  }

  /* Get Index color */
  Get_Keyword_Value(file, "Index_color", value);
  color->value.indexed.index = (PEXTableIndex)atoi(value);

  /* Get Do Color RGB */
  Get_Keyword_Value(file, "Do_color_RGB", value);
  if(strcasecmp(value, "True") == 0)
  {
    color->type = PEXColorTypeRGB;
    set = 1;
  }

  if(set == 0)
  {
    sprintf(errstr, 
      "Color Specifier for '%s' may not have correct type", keyword);
    Warning(errstr);
  }

  /* Get Red */
  Get_Keyword_Value(file, "Red", value);
  color->value.rgb.red = atof(value);

  /* Get Green */
  Get_Keyword_Value(file, "Green", value);
  color->value.rgb.green = atof(value);

  /* Get Blue */
  Get_Keyword_Value(file, "Blue", value);
  color->value.rgb.blue = atof(value);
}

void Load_Lighting_Information(char **file, Input_Parameters *in)
/*
 ******************************************************************************
 **    Description:
 **        This function loads all of the Lighting information
 **        from the parameter file.
 **
 **    Author:
 **       Brad Saunders                Dec. 17, 1993
 **
 **    Input parameters:
 **        char *file - a pointer to the input file.
 **        Input_Parameters in - stores the information being read.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  char value[INPUTSIZE];

  /* Get Lighting index value */
  Get_Keyword_Value(file, "Use_light_entry", value);
  in->light_index = atoi(value);

  in->dcue.mode = PEXOn;
  /* Get Depth Cueing Information */

  /* Get Front Plane */
  Get_Keyword_Value(file, "Front_plane", value);
  in->dcue.front_plane = atof(value);

  /* Get Back Plane */
  Get_Keyword_Value(file, "Back_plane", value);
  in->dcue.back_plane = atof(value);

  /* Get Front Scaling */
  Get_Keyword_Value(file, "Front_scaling", value);
  in->dcue.front_scaling = atof(value);

  /* Get Back Scaling */
  Get_Keyword_Value(file, "Back_scaling", value);
  in->dcue.back_scaling = atof(value);

  Get_Color_Specifier(file, &in->dcue.color, "Depth Cueing");
}

void Load_Rendering_Information(char **file, Input_Parameters *in)
/*
 ******************************************************************************
 **    Description:
 **        This function loads all of the Rendering information
 **        from the parameter file.
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char *file - a pointer to the input file.
 **        Input_Parameters in - stores the information being read.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        file has a valid input file.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  char value[INPUTSIZE];

  /* Get shape hint */
  Get_Keyword_Value(file, "Shape_hint", value);
  in->shape_hint = Convert_Constant(value);

  /* Get ignore edges */
  Get_Keyword_Value(file, "Ignore_edges", value);
  in->ignore_edges = Convert_Flags(value, "Ignore_edges");

  /* Get contour hint */
  Get_Keyword_Value(file, "Contour_hint", value);
  in->contour_hint = Convert_Constant(value);
}

PEXOCRequestType Get_Request_Type(char *value)
/*
 ******************************************************************************
 **    Description:
 **        This function translates the request type string in to
 **        an actual integer value.
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char *value - the request type string.
 **
 **    Output parameters:
 **        PEXOCRequestType the converted request type.
 **
 **    Entry Conditions:
 **        None.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  char errstr[ERRORSTRINGSIZE];

  if(strcmp("PEXOCRender", value) == 0)
    return(PEXOCRender);
  else
    if(strcmp("PEXOCStore", value) == 0)
      return(PEXOCStore);
    else
    {
      sprintf(errstr,"Incorrect Request type - '%s'",value);
      Error(errstr);
    }
}

void Parse_File(char *filename, Input_Parameters *in)
/*
 ******************************************************************************
 **    Description:
 **        This function performs the parsing of the parameter file. 
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char *filename - a pointer to a parameter filename.
 **        Input_Parameters *in - the input structure to be filled.
 **
 **    Output parameters:
 **        None.
 **
 **    Entry Conditions:
 **        None.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        main().
 **
 **    Procedure/Function dependencies:
 **        None.
 **
 ******************************************************************************
 */
{
  char *file;
  char *fileholder;


  init_string_table();

  file = Clean_File(filename);
  fileholder = file;

  Load_Do_Flags(&file, in);
  Load_Initial_Information(&file, in);
  Load_Texture_Information(&file, in);
  Load_Object_Information(&file, in);
  Load_OC_Information(&file, in);
  Load_Binding_LUT_Information(&file, in);
  Load_CoordSource_LUT_Information(&file, in);
  Load_Composition_LUT_Information(&file, in);
  Load_Sampling_LUT_Information(&file, in);
  Load_Viewport_Information(&file, in);
  Load_Lighting_Information(&file, in);
  Load_Rendering_Information(&file, in);

  free(fileholder);
}

char *Clean_File(char *filename)
/*
 ******************************************************************************
 **    Description:
 **        This function reads the parameter file in to a string and 
 **        calls functions to remove spaces and comments from the file.
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char *filename - parameter filename.
 **
 **    Output parameters:
 **        char * - a pointer to the "clean" parameter file string.
 **
 **    Entry Conditions:
 **        None.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Parse_File().
 **
 **    Procedure/Function dependencies:
 **        Parse_File().
 **
 ******************************************************************************
 */
{
  FILE *fp;
  int filesize;			/* size of file in bytes */
  char *file;			/*file in raw form */
  char errstr[ERRORSTRINGSIZE];

  if ((fp = fopen(filename, "r")) == NULL)
  {
    sprintf(errstr,"Cannot open parameter file - '%s'",filename);
    Error(errstr);
  }
  
  fseek(fp, 0L, SEEK_END);   /* crude method for getting size in bytes */
  filesize = ftell(fp);
  fseek(fp, 0L, SEEK_SET);
  file = (char *)malloc(filesize + 1);
  fread(file, 1, filesize, fp);
  file[filesize] = '\0';

  Remove_Spaces(file);		/* shrinks all white space to one space */
  Remove_Comment(file);		/* removes all comments from parfile  */
  return(file);
}

/*  Call any of these functions with a string containing the file  */

void Remove_Spaces(char *incoming)
/*
 ******************************************************************************
 **    Description:
 **        This function removes the spaces and blank lines in the file.
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char *incoming - a string containing the parameter file.
 **
 **    Output parameters:
 **        char * - a pointer to the "clean" parameter file string.
 **
 **    Entry Conditions:
 **        None.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Clean_File().
 **
 **    Procedure/Function dependencies:
 **        Clean_File().
 **
 ******************************************************************************
 */
{
  char ch;
  int space = 0, end_of_line = 0;
  int write_point = 0, read_point = 0, end;
  
  end = strlen(incoming);
  ch = incoming[read_point++];
  while(read_point <= end)
  {
   /*
    * Always write out new lines. 
    * Only put one space between strings.
    * Replace tabs with a space.
    */
    switch(ch)
    {
    case ' ':
      if(!space && !end_of_line)
      {
        space = 1;
        incoming[write_point++] = ch;
      }
      break;
    case '\n':
      if(!end_of_line)
      {
        end_of_line = 1;
        if(space)
          write_point--;
        incoming[write_point++] = ch;
      }
      break; 
    case '\t':
      if(!space && incoming[write_point-1] != '\n')
      {
        space = 1;
        incoming[write_point++] = ' ';
      }
      break;
    default:
      space = 0;
      end_of_line = 0;
      incoming[write_point++] = ch;
      break;
    }
    ch = incoming[read_point++];
  }
  incoming[write_point] = '\0';
}  

void Remove_Comment(char *incoming)
/*
 ******************************************************************************
 **    Description:
 **        This function removes all of the comments from the paramter
 **        file string.  A comment is demoted by the  value in 
 **        COMMENT_DELIMITER.
 **
 **    Author:
 **       Brad Saunders                Oct. 18, 1993
 **
 **    Input parameters:
 **        char *incoming - a string containing the parameter file.
 **
 **    Output parameters:
 **        char * - a pointer to the "clean" parameter file string.
 **
 **    Entry Conditions:
 **        None.
 **
 **    Exit Conditions:
 **        None.
 **
 **    Error Conditions:
 **        None.
 **
 **    Referenced by:
 **        Clean_File().
 **
 **    Procedure/Function dependencies:
 **        Clean_File().
 **
 ******************************************************************************
 */
{
  char ch;
  int write_point = 0, read_point = 0, end;
    
  end = strlen(incoming);
  ch = incoming[read_point++];
  while(read_point <= end)
  {
    if(ch == COMMENT_DELIMITER)
    {
      if(write_point > 0)
        write_point--;
      while(incoming[write_point] == 32)
        write_point--;
      if(write_point != 0)
        write_point++;
      
      while(incoming[read_point] != '\n' && read_point <= end)
        read_point++;
      read_point++;
      if(incoming[write_point-1] != '\n' && write_point != 0)
        incoming[write_point++] = '\n';     
    }
    else
      incoming[write_point++] = ch;
    ch = incoming[read_point++];
  }
  incoming[write_point] = '\0';
}
